from __future__ import annotations

import asyncio
import socket
import logging
from collections import deque
from typing import Any, AsyncGenerator

from .parser import SIPMessage, SIPParser, SIPRequest


class SIPProtocol(asyncio.DatagramProtocol):
    def __init__(self, client: SIPClient) -> None:
        self.client = client

    def datagram_received(self, data: bytes, _: tuple[str | Any, int]) -> None:
        self.client._on_datagram(data)

    def connection_lost(self, exc: Exception | None) -> None:
        self.client._close_connection(exc)


class SIPClient(AsyncGenerator[SIPMessage, None]):
    transport: asyncio.DatagramTransport | None
    closed_event: asyncio.Event
    queue: deque[SIPMessage]
    logger: logging.Logger

    def __init__(
        self,
        host: str,
        port: int,
        parser: SIPParser,
        logger: logging.Logger = logging.getLogger(__name__),
    ) -> None:
        self.logger = logger
        self.port = port
        self.host = host
        self.parser = parser
        self.queue = deque()
        self.transport = None
        self.new_msg_event = asyncio.Event()
        self.closed_event = asyncio.Event()

    async def send_message(self, msg: SIPRequest) -> None:
        assert self.transport is not None, "Client is not connected"
        content = msg.serialize()
        self.logger.debug(
            f"(send_message) Sending content: {content.decode()}"
        )
        self.transport.sendto(data=content)

    async def connect(self, loop: asyncio.AbstractEventLoop) -> None:
        if self.transport is not None:
            self.logger.debug("(connect) Already connected...")
            return
        self.logger.debug(
            "(connect) Creating datagram endpoint to "
            f"{self.host}:{self.port}"
        )
        self.transport, _ = await loop.create_datagram_endpoint(
            lambda: SIPProtocol(self),
            remote_addr=(self.host, self.port),
            family=socket.AF_INET,
        )
        self.logger.info(
            f"(connect) Created datagram endpoint to {self.host}:{self.port}"
        )
        self.closed_event.clear()

    def get_message(self) -> SIPMessage | None:
        if len(self.queue) == 0:
            return None
        return self.queue.popleft()

    async def wait_for_message(self) -> SIPMessage | None:
        msg = self.get_message()
        if msg is not None:
            self.logger.debug(f"(wait_for_message) returning msg: {msg}")
            return msg
        self.logger.debug("(wait_for_message) blocking on event...")
        await self.new_msg_event.wait()
        self.new_msg_event.clear()
        return self.get_message()

    def __aiter__(self) -> SIPClient:
        return self

    async def asend(self, _: None) -> SIPMessage:
        msg = await self.wait_for_message()
        if msg is None:
            raise StopAsyncIteration()
        return msg

    async def athrow(self, *args: Any, **kwargs: Any) -> SIPMessage:
        return await super().athrow(*args, **kwargs)

    async def disconnect(self) -> None:
        if self.transport is None:
            self.logger.debug("(disconnect) Not connected, skipping...")
            return
        if not self.transport.is_closing():
            self.transport.close()
        self.logger.info(
            f"(disconnect) Disconnected from {self.host}:{self.port}"
        )
        await self.closed_event.wait()
        self.closed_event.clear()

    def clear_queue(self) -> None:
        self.queue.clear()

    def _on_datagram(self, buffer: bytes) -> None:
        self.logger.debug(f"Received datagram: \n{buffer.decode()}")
        try:
            msg = self.parser.parse(buffer.decode())
            if isinstance(msg, Exception):
                raise msg
            self.queue.append(msg)
            self.new_msg_event.set()
        except BaseException:
            self.logger.exception("Parsing error: ")
            raise

    def _close_connection(self, exc: Exception | None) -> None:
        if exc is not None:
            self.logger.error(f"Error on close_connection: {exc}")
        self.closed_event.set()
        self.new_msg_event.set()

    def get_local_addr(self) -> tuple[str, int]:
        assert self.transport is not None
        host, port = self.transport.get_extra_info("sockname")
        return host, port



from __future__ import annotations
import logging

import re
from dataclasses import dataclass

from .utils import (
    build_contact,
    build_digest_response,
    build_from,
    build_request_uri,
    build_to,
    build_via,
    gen_branch,
    get_host_and_username_from_contact,
)

digest_regex = re.compile(r'(\w+)[:=][\s"]?([^",]+)"?')


@dataclass
class SIPBaseMessage:
    version: str
    body: str | None
    headers: dict[str, str]


@dataclass
class SIPRequest(SIPBaseMessage):
    method: str
    request_uri: str

    def __post_init__(self) -> None:
        self.headers["Content-Length"] = str(len(self.body or ""))

    def serialize(self) -> bytes:
        buffer = f"{self.method} {self.request_uri} {self.version}\r\n"
        headers = self.headers
        for key, value in headers.items():
            buffer += f"{key}: {value}\r\n"
        buffer += "\r\n"
        if self.body is not None:
            buffer += self.body
        return buffer.encode()


@dataclass
class SIPRegisterRequest(SIPRequest):
    @classmethod
    def build_new(
        cls,
        username: str,
        local_host: str,
        local_port: int,
        host: str,
        port: int,
        call_id: str,
        expires_s: int,
        cseq: int = 1,
    ) -> SIPRegisterRequest:
        local_contact = build_contact(username, local_host, local_port)
        remote_contact = build_contact(username, host, port)
        headers = {
            "Contact": local_contact,
            "Via": build_via(local_host, local_port),
            "CSeq": f"{cseq} REGISTER",
            "Call-ID": call_id,
            "From": build_from(username, remote_contact),
            "To": build_to(username, remote_contact),
            "Max-Forwards": "70",
            "Expires": str(expires_s),
        }
        if expires_s is not None:
            headers["Expires"] = str(expires_s)
        return SIPRegisterRequest(
            method="REGISTER",
            request_uri=build_request_uri(username, host, port),
            version="SIP/2.0",
            headers=headers,
            body=None,
        )

    @classmethod
    def build_from_response(
        cls, request: SIPRegisterRequest, response: SIPResponse, password: str
    ) -> SIPRegisterRequest:
        username, host = get_host_and_username_from_contact(
            request.headers["Contact"]
        )
        authenticate = response.headers["WWW-Authenticate"]
        options = dict(digest_regex.findall(authenticate))
        auth_response = build_digest_response(
            realm=options["realm"],
            method=request.method,
            host=host,
            username=username,
            password=password,
            nonce=options["nonce"],
            opaque=options["opaque"],
        )
        return SIPRegisterRequest(
            method="REGISTER",
            request_uri=request.request_uri,
            version="SIP/2.0",
            headers={
                **request.headers,
                "Via": response.headers["Via"].split(";")[0]
                + f";branch={gen_branch()};alias",
                "Authorization": auth_response,
            },
            body=None,
        )


@dataclass
class SIPResponse(SIPBaseMessage):
    status: int
    reason: str


SIPMessage = SIPRequest | SIPResponse


class SIPParser:
    def __init__(
        self, logger: logging.Logger = logging.getLogger(__name__)
    ) -> None:
        self.logger = logger

    def parse_request(self, message: str) -> SIPRequest | Exception:
        self.logger.debug(f"(parse_request) message: {message}")
        lines = message.split("\r\n")
        self.logger.debug(f"(parse_request) lines: {lines}")
        start_line = lines.pop(0)
        self.logger.debug(f"(parse_request) start_line: {start_line}")
        method, request_uri, version = start_line.split(" ")
        result = self.parse_headers_and_body(lines)
        if isinstance(result, ValueError):
            return result
        headers, body = result
        return SIPRequest(
            body=body,
            method=method,
            version=version,
            headers=headers,
            request_uri=request_uri,
        )

    def parse_headers_and_body(
        self, lines: list[str]
    ) -> tuple[dict[str, str], str | None] | ValueError:
        headers: dict[str, str] = {}
        line_break_index: int | None = None
        for index, line in enumerate(lines):
            if line == "":
                line_break_index = index
                break
            self.logger.debug(
                f"(parse_headers_and_body) line before split: {line}"
            )
            key, value = line.split(":", 1)
            headers[key.strip()] = value.strip()
        body: str | None = None
        if line_break_index is not None and len(lines) > line_break_index + 1:
            body = "\r\n".join(lines[line_break_index + 1 :])
        return headers, body

    def parse_response(self, message: str) -> SIPResponse | Exception:
        try:
            lines = message.split("\r\n")
            status_line = lines.pop(0)
            version, status, reason = status_line.split(" ")
            result = self.parse_headers_and_body(lines)
            if isinstance(result, ValueError):
                return result
            headers, body = result
            return SIPResponse(
                body=body,
                reason=reason,
                version=version,
                headers=headers,
                status=int(status),
            )
        except Exception as exc:
            return exc

    def parse(self, buffer: str) -> SIPMessage | Exception:
        result: SIPMessage | Exception
        result = self.parse_response(buffer)
        if isinstance(result, SIPResponse):
            return result
        result = self.parse_request(buffer)
        return result



import hashlib
import uuid


def build_request_uri(username: str, host: str, port: int) -> str:
    return f"sip:{username}@{host}:{port}"


def gen_branch() -> str:
    return "z9hG4bK." + uuid.uuid4().hex[:24]


def build_contact(username: str, host: str, port: int) -> str:
    return f"<sip:{username}@{host}:{port}>"


def get_host_and_username_from_contact(contact: str) -> tuple[str, str]:
    username, host = contact[1:-1].split(":", 1)[1].split("@")
    return username, host


def build_via(host: str, port: int) -> str:
    return f"SIP/2.0/UDP {host}:{port};" f"branch={gen_branch()};alias"


def build_from(username: str, contact: str) -> str:
    return f'"{username}" {contact}'


def build_to(username: str, contact: str) -> str:
    return f'"{username}" {contact}'


def build_digest(
    method: str,
    host: str,
    realm: str,
    nonce: str,
    username: str,
    password: str,
) -> str:
    part1 = hashlib.md5(f"{username}:{realm}:{password}".encode()).hexdigest()
    part2 = hashlib.md5(f"{method}:{host}".encode()).hexdigest()
    return hashlib.md5(f"{part1}:{nonce}:{part2}".encode()).hexdigest()


def build_digest_response(
    method: str,
    host: str,
    realm: str,
    nonce: str,
    username: str,
    password: str,
    opaque: str,
) -> str:
    hash = build_digest(
        method=method,
        host=host,
        realm=realm,
        nonce=nonce,
        username=username,
        password=password,
    )
    return (
        f'Digest username="{username}", realm="{realm}", '
        f'nonce="{nonce}", uri="{host}", algorithm=MD5, '
        f'opaque="{opaque}", response="{hash}"'
    )


def gen_call_id() -> str:
    return str(uuid.uuid4())



import asyncio
import logging

from .client import SIPClient
from .parser import SIPRegisterRequest, SIPRequest, SIPResponse
from .utils import gen_call_id


class SIPHolder:
    task: asyncio.Task[None] | None

    def __init__(
        self,
        client: SIPClient,
        username: str,
        password: str,
        registration_expires_s: int = 30,
        logger: logging.Logger = logging.getLogger(__name__),
    ) -> None:
        self.logger = logger
        self.client = client
        self.registration_count = 1
        self.should_stop = False
        self.task = None
        self.username = username
        self.password = password
        self.call_id = gen_call_id()
        self.registration_expires_s = registration_expires_s

    async def start(self) -> None:
        self.task = asyncio.create_task(self.task_main())

    async def get_task_error(self) -> Exception | None:
        if self.task is None:
            return None
        if not self.task.done():
            return None
        try:
            self.task.result()
            return None
        except Exception as exc:
            return exc

    async def task_main(self) -> None:
        while not self.should_stop:
            await self.register()
            self.logger.debug(f"Sleeping for {self.registration_expires_s}s")
            await asyncio.sleep(self.registration_expires_s)

    async def register(self) -> None:
        self.logger.debug("Starting register routine")
        local_host, local_port = self.client.get_local_addr()
        self.logger.debug(f"Local addr: {local_host}:{local_port}")
        request = SIPRegisterRequest.build_new(
            username=self.username,
            host=self.client.host,
            port=self.client.port,
            local_host=local_host,
            local_port=local_port,
            call_id=self.call_id,
            cseq=self.registration_count,
            expires_s=self.registration_expires_s,
        )
        self.logger.debug(
            f"Builded new register request: \n {request.serialize().decode()}"
        )
        response = await self.send_and_receive(request)
        if isinstance(response, Exception):
            raise response
        request_with_auth = SIPRegisterRequest.build_from_response(
            request=request, response=response, password=self.password
        )
        response = await self.send_and_receive(request_with_auth)
        if isinstance(response, Exception):
            raise response
        self.registration_count += 1

    async def send_and_receive(
        self, request: SIPRequest
    ) -> SIPResponse | Exception:
        self.client.clear_queue()
        last_error: Exception | None = None
        for _ in range(0, 3):
            await self.client.send_message(request)
            try:
                message = await asyncio.wait_for(
                    self.client.wait_for_message(), timeout=10
                )
                self.logger.debug(f"Received message: {message}")
                if message is None:
                    last_error = RuntimeError(
                        "Returned None when expected response"
                    )
                elif isinstance(message, SIPRequest):
                    last_error = RuntimeError(
                        f"Received request: {message} when expected response"
                    )
                else:
                    return message
            except asyncio.TimeoutError:
                last_error = RuntimeError("Waiting for message timed out")
                continue
        assert last_error is not None
        return last_error

    async def stop(self) -> None:
        if self.task is None:
            return
        self.should_stop = True
        try:
            await asyncio.wait_for(self.task, timeout=3)
            return
        except asyncio.TimeoutError:
            pass
        except BaseException:
            self.logger.exception("Exception from stopped task: ")
            return
        self.task.cancel()
        try:
            await self.task
        except asyncio.CancelledError:
            pass



import os
from pathlib import Path

from setuptools import find_packages, setup  # type: ignore [import-untyped]


setup(
    name="isip-client",
    version=os.environ["GITHUB_REF_NAME"],
    author="Vladimir Vojtenko",
    author_email="vladimirdev635@gmail.com",
    license="MIT",
    packages=find_packages(exclude=["__tests__*"]),
    include_package_data=True,
    long_description=(Path(__file__).parent / "README.md").read_text(),
    long_description_content_type="text/markdown",
)
